#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "body.hh"

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
Body<GridTEMPLATE, VectorTEMPLATE, dims>::Body(
        const std::shared_ptr<BodyData<dims>>& bodyData,
        const std::shared_ptr<GridType>& grid) :
    bodyData_(bodyData),
    grid_(grid),
    deformation_(std::make_shared<DeformationFunction>(grid_->leafGridView())),
    deformedGrid_(std::make_shared<DeformedGridType>(*grid_, *deformation_)),
    leafView_(deformedGrid_->leafGridView()),
    leafVertexCount_(leafView_.size(dims)),
    assembler_(std::make_shared<Assembler>(leafView_)),
    matrices_()
{
    leafBoundaryConditions_.resize(0);

    levelBoundaryConditions_.resize(grid_->maxLevel()+1);
    for (size_t i=0; i<levelBoundaryConditions_.size(); i++) {
        levelBoundaryConditions_[i].resize(0);
    }

    externalForce_ = [&](double relativeTime, VectorType &ell) {
        // Problem formulation: right-hand side
        std::vector<std::shared_ptr<LeafBoundaryCondition>> leafNeumannConditions;
        leafBoundaryConditions("neumann", leafNeumannConditions);
        ell.resize(gravityFunctional_.size());
        ell = gravityFunctional_;

        for (size_t i=0; i<leafNeumannConditions.size(); i++) {
            const auto& leafNeumannCondition = leafNeumannConditions[i];

            VectorType b;
            assembler_->assembleNeumann(*leafNeumannCondition->boundaryPatch(), b, *leafNeumannCondition->boundaryFunction(),
                                    relativeTime);
            ell += b;
        }
    };
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::assemble() {

    // assemble matrices
    assembler_->assembleElasticity(bodyData_->getYoungModulus(), bodyData_->getPoissonRatio(), *matrices_.elasticity);
    assembler_->assembleViscosity(bodyData_->getShearViscosityField(), bodyData_->getBulkViscosityField(), *matrices_.damping);
    assembler_->assembleMass(bodyData_->getDensityField(), *matrices_.mass);

    // assemble state energy norm
    typename FrictionBoundaryCondition::BoundaryPatch frictionBoundary(this->leafView(), false); // boundary patch containing all friction boundaries (possibly disconnected)
    for (size_t i=0; i<frictionBoundaryConditions_.size(); i++) {
        const auto& frictionCondition = frictionBoundaryConditions_[i];
        //std::cout << "grid Pointer " << frictionCondition->boundaryPatch()-> << std::endl;
        frictionBoundary.addPatch(*frictionCondition->boundaryPatch());
    }

    ScalarMatrix relativeFrictionBoundaryMass;
    assembler_->assembleFrictionalBoundaryMass(frictionBoundary, relativeFrictionBoundaryMass);
    relativeFrictionBoundaryMass /= frictionBoundary.area(); // TODO: weight by individual friction patches?
    stateEnergyNorm_ = std::make_shared<const EnergyNorm<ScalarMatrix, ScalarVector>>(relativeFrictionBoundaryMass);

    // assemble forces
    assembler_->assembleBodyForce(bodyData_->getGravityField(), gravityFunctional_);
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::assembleFriction(
        const Config::FrictionModel& frictionModel,
        const ScalarVector& weightedNormalStress) {

    globalFriction_.resize({frictionBoundaryConditions_.size()});

    for (size_t i=0; i<globalFriction_.size(); i++) {
        const auto& frictionBoundaryCondition = frictionBoundaryConditions_[i];
        globalFriction_[i] = assembler_->assembleFrictionNonlinearity(frictionModel, *frictionBoundaryCondition->boundaryPatch(), *frictionBoundaryCondition->frictionData(), weightedNormalStress);
    }
}

// setter and getter

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::data() const
-> const std::shared_ptr<BodyData<dims>>& {

    return bodyData_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::setDeformation(const VectorType& def) {
    deformation_->setDeformation(def);
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::deformation() const
-> DeformationFunction& {

    return *deformation_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::deformedGrid() const
-> const DeformedGridType* {

    return deformedGrid_.get();
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::leafView() const
-> const DeformedLeafGridView& {

    return leafView_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::leafVertexCount() const
-> int {

    return leafVertexCount_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::assembler() const
-> const std::shared_ptr<Assembler>& {

    return assembler_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::matrices() const
-> const Matrices<Matrix,1>& {

    return matrices_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::externalForce() const
-> const ExternalForce& {

    return externalForce_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::stateEnergyNorm() const
-> const std::shared_ptr<const StateEnergyNorm>& {

    return stateEnergyNorm_;
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::addBoundaryCondition(std::shared_ptr<LeafBoundaryCondition> leafBoundaryCondition) {
    leafBoundaryConditions_.emplace_back(leafBoundaryCondition);
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::leafBoundaryConditions(
        const std::string& tag,
        std::vector<std::shared_ptr<LeafBoundaryCondition>>& selectedConditions) const {

    if (tag == "friction") {
        selectedConditions.resize(frictionBoundaryConditions_.size());
        for (size_t i=0; i<frictionBoundaryConditions_.size(); i++) {
                    selectedConditions[i] = frictionBoundaryConditions_[i];
        }

    } else {
        selectedConditions.resize(0);
        for (size_t i=0; i<leafBoundaryConditions_.size(); i++) {
            if (leafBoundaryConditions_[i]->tag() == tag)
                selectedConditions.emplace_back(leafBoundaryConditions_[i]);
        }
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::leafBoundaryConditions() const
-> const std::vector<std::shared_ptr<LeafBoundaryCondition>>& {

    return leafBoundaryConditions_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::addBoundaryCondition(
        std::shared_ptr<LevelBoundaryCondition> levelBoundaryCondition,
        int level) {

    levelBoundaryConditions_[level].push_back(levelBoundaryCondition);
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::levelBoundaryConditions(
        const std::string& tag,
        std::vector<std::vector<std::shared_ptr<LevelBoundaryCondition>>>& selectedConditions) const {

    selectedConditions.resize(levelBoundaryConditions_.size());
    for (size_t level=0; level<levelBoundaryConditions_.size(); level++) {
        const std::vector<std::shared_ptr<LevelBoundaryCondition>>& levelBoundaryConditions = levelBoundaryConditions_[level];
        for (size_t i=0; i<levelBoundaryConditions.size(); i++) {
            if (levelBoundaryConditions[i]->tag() == tag)
                selectedConditions[level].push_back(levelBoundaryConditions[i]);
        }
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::levelBoundaryConditions(
        const std::string& tag,
        int level,
        std::vector<std::shared_ptr<LevelBoundaryCondition>>& selectedConditions) const {

    selectedConditions.resize(0);
    const std::vector<std::shared_ptr<LevelBoundaryCondition>>& levelBoundaryConditions = levelBoundaryConditions_[level];
    for (size_t i=0; i<levelBoundaryConditions.size(); i++) {
        if (levelBoundaryConditions[i]->tag() == tag)
            selectedConditions.push_back(levelBoundaryConditions[i]);
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::levelBoundaryConditions(int level) const
-> const std::vector<std::shared_ptr<LevelBoundaryCondition>>& {

    return levelBoundaryConditions_[level];
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::levelBoundaryConditions() const
-> const std::vector<std::vector<std::shared_ptr<LevelBoundaryCondition>>>& {

    return levelBoundaryConditions_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::addBoundaryCondition(std::shared_ptr<FrictionBoundaryCondition> frictionBoundaryCondition) {
    frictionBoundaryConditions_.emplace_back(frictionBoundaryCondition);
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::frictionBoundaryConditions() const
-> const std::vector<std::shared_ptr<FrictionBoundaryCondition>>& {

    return frictionBoundaryConditions_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
auto Body<GridTEMPLATE, VectorTEMPLATE, dims>::globalFriction()
-> GlobalFrictionContainer& {
    return globalFriction_;
}


template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::boundaryPatchNodes(
        const std::string& tag,
        BoundaryPatchNodes& nodes) const {

    std::vector<std::shared_ptr<typename Body::LeafBoundaryCondition>> boundaryConditions;
    this->leafBoundaryConditions(tag, boundaryConditions);

    nodes.resize(boundaryConditions.size());

    for (size_t i=0; i<nodes.size(); i++) {
        nodes[i] = boundaryConditions[i]->boundaryPatch()->getVertices();
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::boundaryNodes(
        const std::string& tag,
        BoundaryNodes& nodes) const {

    std::vector<std::shared_ptr<typename Body::LeafBoundaryCondition>> boundaryConditions;
    this->leafBoundaryConditions(tag, boundaryConditions);

    nodes.resize(boundaryConditions.size());

    for (size_t i=0; i<nodes.size(); i++) {
        nodes[i] = boundaryConditions[i]->boundaryNodes().get();
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::boundaryFunctions(
        const std::string& tag,
        BoundaryFunctions& functions) const {

    std::vector<std::shared_ptr<typename Body::LeafBoundaryCondition>> boundaryConditions;
    this->leafBoundaryConditions(tag, boundaryConditions);

    functions.resize(boundaryConditions.size());

    for (size_t i=0; i<functions.size(); i++) {
        functions[i] = boundaryConditions[i]->boundaryFunction().get();
    }
}

template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
void Body<GridTEMPLATE, VectorTEMPLATE, dims>::boundaryPatches(
        const std::string& tag,
        BoundaryPatches& patches) const {

    std::vector<std::shared_ptr<typename Body::LeafBoundaryCondition>> boundaryConditions;
    this->leafBoundaryConditions(tag, boundaryConditions);

    patches.resize(boundaryConditions.size());

    for (size_t i=0; i<patches.size(); i++) {
        patches[i] = boundaryConditions[i]->boundaryPatch().get();
    }
}

#include "body_tmpl.cc"
