#ifndef SRC_DATA_STRUCTURES_BODY_HH
#define SRC_DATA_STRUCTURES_BODY_HH

#include <functional>

#include <dune/common/bitsetvector.hh>
#include <dune/common/function.hh>
#include <dune/common/shared_ptr.hh>

#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

//#include <dune/fufem/assemblers/assembler.hh>
//#pragma clang diagnostic push
//#pragma clang diagnostic ignored "-Wsign-compare"
//#include <dune/fufem/functionspacebases/p0basis.hh>
//#pragma clang diagnostic pop
//#include <dune/fufem/functionspacebases/p1nodalbasis.hh>

#include <dune/grid/geometrygrid/grid.hh>

#include <dune/fufem/functions/deformationfunction.hh>

#include <dune/solvers/norms/energynorm.hh>

#include <dune/tectonic/bodydata.hh>
#include <dune/tectonic/globalfriction.hh>
#include <dune/tectonic/globalfrictiondata.hh>

#include "../assemblers.hh"
#include "../boundarycondition.hh"
#include "globalfrictioncontainer.hh"
#include "enums.hh"
#include "matrices.hh"



template <class GridTEMPLATE, class VectorTEMPLATE, int dims>
class Body {
public:
    using GridType = GridTEMPLATE;
    using VectorType = VectorTEMPLATE;

    using DeformationFunction = DeformationFunction<typename GridType::LeafGridView, VectorType>;
    using DeformedGridType = Dune::GeometryGrid<GridType, DeformationFunction>;
    using DeformedLeafGridView = typename DeformedGridType::LeafGridView;
    using DeformedLevelGridView = typename DeformedGridType::LevelGridView;

    using Assembler = MyAssembler<DeformedLeafGridView, dims>;
    using Matrix = typename Assembler::Matrix;
    using LocalVector = typename VectorType::block_type;

    using ScalarMatrix = Dune::BCRSMatrix<Dune::FieldMatrix<double, 1, 1>>;
    using ScalarVector = Dune::BlockVector<Dune::FieldVector<double, 1>>;

    using Function = Dune::VirtualFunction<double, double>;
    using ExternalForce = std::function<void(double, VectorType &)>;

    using LeafBoundaryCondition = BoundaryCondition<DeformedLeafGridView, dims>;
    using FrictionBoundaryCondition = FrictionBoundaryCondition<DeformedLeafGridView, LocalVector, dims>;
    using LevelBoundaryCondition = BoundaryCondition<DeformedLevelGridView, dims>;

    using BoundaryFunctions = std::vector<const Function* >;
    using BoundaryNodes = std::vector<const Dune::BitSetVector<dims>* >;
    using BoundaryPatchNodes = std::vector<const Dune::BitSetVector<1>* >;
    using BoundaryPatches = std::vector<const typename LeafBoundaryCondition::BoundaryPatch* >;

    using GlobalFriction = GlobalFriction<Matrix, VectorType>;
    using GlobalFrictionContainer = GlobalFrictionContainer<GlobalFriction, 1>;

    using StateEnergyNorm = EnergyNorm<ScalarMatrix, ScalarVector>;

private:
    std::shared_ptr<Body<GridType, VectorType, dims>> parent_;
    std::vector<std::shared_ptr<Body<GridType, VectorType, dims>>> children_;

    std::shared_ptr<BodyData<dims>> bodyData_;

    std::shared_ptr<GridType> grid_;
    std::shared_ptr<DeformationFunction> deformation_;
    std::shared_ptr<DeformedGridType> deformedGrid_;

    DeformedLeafGridView leafView_;
    int leafVertexCount_;

    std::shared_ptr<Assembler> assembler_;
    Matrices<Matrix,1> matrices_;

    ExternalForce externalForce_;

    VectorType gravityFunctional_;
    std::shared_ptr<const StateEnergyNorm> stateEnergyNorm_;

    // boundary conditions
    std::vector<std::shared_ptr<LeafBoundaryCondition>> leafBoundaryConditions_;
    std::vector<std::vector<std::shared_ptr<LevelBoundaryCondition>>> levelBoundaryConditions_; // first index: level, second index: bc on lvl

    // friction boundary conditions
    std::vector<std::shared_ptr<FrictionBoundaryCondition>> frictionBoundaryConditions_;
    GlobalFrictionContainer globalFriction_;

public:
    Body(
            const std::shared_ptr<BodyData<dims>>& bodyData,
            const std::shared_ptr<GridType>& grid);

    void assemble();
    void assembleFriction(
            const Config::FrictionModel& frictionModel,
            const ScalarVector& weightedNormalStress);

    // setter and getter
    auto data() const -> const std::shared_ptr<BodyData<dims>>&;

    void setDeformation(const VectorType& def);
    auto deformation() const -> DeformationFunction&;
    auto deformedGrid() const -> const DeformedGridType*;

    auto leafView() const -> const DeformedLeafGridView&;
    auto leafVertexCount() const -> int;

    auto assembler() const -> const std::shared_ptr<Assembler>&;
    auto matrices() const -> const Matrices<Matrix,1>&;

    auto externalForce() const -> const ExternalForce&;

    auto stateEnergyNorm() const -> const std::shared_ptr<const StateEnergyNorm>&;

    void addBoundaryCondition(std::shared_ptr<LeafBoundaryCondition> leafBoundaryCondition);
    void leafBoundaryConditions(
            const std::string& tag,
            std::vector<std::shared_ptr<LeafBoundaryCondition>>& selectedConditions) const;
    auto leafBoundaryConditions() const -> const std::vector<std::shared_ptr<LeafBoundaryCondition>>&;

    void addBoundaryCondition(std::shared_ptr<LevelBoundaryCondition> levelBoundaryCondition, int level);
    void levelBoundaryConditions(
            const std::string& tag,
            std::vector<std::vector<std::shared_ptr<LevelBoundaryCondition>>>& selectedConditions) const;
    void levelBoundaryConditions(
            const std::string& tag,
            int level,
            std::vector<std::shared_ptr<LevelBoundaryCondition>>& selectedConditions) const;
    auto levelBoundaryConditions(int level) const -> const std::vector<std::shared_ptr<LevelBoundaryCondition>>&;
    auto levelBoundaryConditions() const -> const std::vector<std::vector<std::shared_ptr<LevelBoundaryCondition>>>&;

    void addBoundaryCondition(std::shared_ptr<FrictionBoundaryCondition> frictionBoundaryCondition);
    auto frictionBoundaryConditions() const -> const std::vector<std::shared_ptr<FrictionBoundaryCondition>>&;

    auto globalFriction() -> GlobalFrictionContainer&;

    void boundaryPatchNodes(
            const std::string& tag,
            BoundaryPatchNodes& nodes) const;
    void boundaryNodes(
            const std::string& tag,
            BoundaryNodes& nodes) const;
    void boundaryFunctions(
            const std::string& tag,
            BoundaryFunctions& functions) const;
    void boundaryPatches(
            const std::string& tag,
            BoundaryPatches& patches) const;
};
#endif
