#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/contact/projections/normalprojection.hh>

#include "levelcontactnetwork.hh"

#include "../utils/tobool.hh"
#include "../utils/debugutils.hh"

template <class GridType, int dimension>
LevelContactNetwork<GridType, dimension>::LevelContactNetwork(
        int nBodies,
        int nCouplings,
        int level) :
    level_(level),
    bodies_(nBodies),
    couplings_(nCouplings),
    deformedGrids_(nBodies),
    leafViews_(nBodies),
    levelViews_(nBodies),
    leafVertexCounts_(nBodies),
    nBodyAssembler_(nBodies, nCouplings)
{}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::assemble() {
    for (size_t i=0; i<nBodies(); i++) {
        bodies_[i]->assemble();
        //printBasisDofLocation(bodies_[i]->assembler()->vertexBasis); //TODO remove after debugging
    }

    // set up dune-contact nBodyAssembler
    nBodyAssembler_.setGrids(deformedGrids_);

    for (size_t i=0; i<nCouplings(); i++) {
      nBodyAssembler_.setCoupling(*couplings_[i], i);
    }

    nBodyAssembler_.assembleTransferOperator();
    nBodyAssembler_.assembleObstacle();
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::assembleFriction(
        const Config::FrictionModel& frictionModel,
        const std::vector<ScalarVector>& weightedNormalStress) {

    for (size_t i=0; i<nBodies(); i++) {
        bodies_[i]->assembleFriction(frictionModel, weightedNormalStress[i]);
    }
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::setBodies(const std::vector<std::shared_ptr<Body>>& bodies) {
    assert(nBodies()==bodies.size());
    bodies_ = bodies;

    matrices_.elasticity.resize(nBodies());
    matrices_.damping.resize(nBodies());
    matrices_.mass.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        deformedGrids_[i] = bodies_[i]->deformedGrid();

        leafViews_[i] = std::make_unique<DeformedLeafGridView>(deformedGrids_[i]->leafGridView());

        levelViews_[i].resize(deformedGrids_[i]->maxLevel()+1);
        for (size_t j=0; j<levelViews_[i].size(); j++) {
            levelViews_[i][j] = std::make_unique<DeformedLevelGridView>(deformedGrids_[i]->levelGridView(j));
        }

        leafVertexCounts_[i] = leafViews_[i]->size(dimension);

        matrices_.elasticity[i] = bodies_[i]->matrices().elasticity;
        matrices_.damping[i] = bodies_[i]->matrices().damping;
        matrices_.mass[i] = bodies_[i]->matrices().mass;
    }
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::setCouplings(const std::vector<std::shared_ptr<CouplingPair>>& couplings) {
    assert(this->nCouplings()==couplings.size());
    couplings_ = couplings;
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::constructBody(
        const std::shared_ptr<BodyData<dimension>>& bodyData,
        const std::shared_ptr<GridType>& grid,
        std::shared_ptr<Body>& body) const {

    body = std::make_shared<Body>(bodyData, grid);
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::constructCoupling(
        int nonMortarBodyIdx, int mortarBodyIdx,
        const std::shared_ptr<DeformedLevelBoundaryPatch>& nonmortarPatch,
        const std::shared_ptr<DeformedLevelBoundaryPatch>& mortarPatch,
        std::shared_ptr<CouplingPair>& coupling) const {

    coupling = std::make_shared<CouplingPair>();

    auto contactProjection = std::make_shared<Dune::Contact::NormalProjection<DeformedLeafBoundaryPatch>>();
    std::shared_ptr<typename CouplingPair::BackEndType> backend = nullptr;
    coupling->set(nonMortarBodyIdx, mortarBodyIdx, nonmortarPatch, mortarPatch, 0.1, Dune::Contact::CouplingPairBase::STICK_SLIP, contactProjection, backend);
}

// collects all leaf boundary nodes from the different bodies identified by "tag" in a single BitSetVector
template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::totalNodes(
        const std::string& tag,
        Dune::BitSetVector<dimension>& totalNodes) const {

    int totalSize = 0;
    for (size_t i=0; i<nBodies(); i++) {
        totalSize += this->body(i)->leafVertexCount();
    }

    totalNodes.resize(totalSize);

    int idx=0;
    for (size_t i=0; i<nBodies(); i++) {
        const auto& body = this->body(i);
        std::vector<std::shared_ptr<typename Body::LeafBoundaryCondition>> boundaryConditions;
        body->leafBoundaryConditions(tag, boundaryConditions);

        const int idxBackup = idx;
        for (size_t bc = 0; bc<boundaryConditions.size(); bc++) {
            const auto& nodes = boundaryConditions[bc]->boundaryNodes();
            for (size_t j=0; j<nodes->size(); j++, idx++)
                for (int k=0; k<dimension; k++)
                    totalNodes[idx][k] = (*nodes)[j][k];
            idx = (bc==boundaryConditions.size()-1 ? idx : idxBackup);
        }
    }
}

// collects all leaf boundary nodes from the different bodies identified by "tag" in a vector of BitSetVector
template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::boundaryPatchNodes(
        const std::string& tag,
        BoundaryPatchNodes& nodes) const {

    nodes.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        this->body(i)->boundaryPatchNodes(tag, nodes[i]);
    }
}

// collects all leaf boundary nodes from the different bodies identified by "tag" in a vector of BitSetVector
template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::boundaryNodes(
        const std::string& tag,
        BoundaryNodes& nodes) const {

    nodes.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        this->body(i)->boundaryNodes(tag, nodes[i]);
    }
}

// collects all leaf boundary functions from the different bodies identified by "tag"
template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::boundaryFunctions(
        const std::string& tag,
        BoundaryFunctions& functions) const {

    functions.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        this->body(i)->boundaryFunctions(tag, functions[i]);
    }
}

// collects all leaf boundary patches from the different bodies identified by "tag"
template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::boundaryPatches(
        const std::string& tag,
        BoundaryPatches& patches) const {

    patches.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        this->body(i)->boundaryPatches(tag, patches[i]);
    }
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::stateEnergyNorms(StateEnergyNorms& stateEnergyNorms) const {
    stateEnergyNorms.resize(nBodies());
    for (size_t i=0; i<stateEnergyNorms.size(); i++) {
        stateEnergyNorms[i] = body(i)->stateEnergyNorm().get();
    }
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::level() const
-> int {

    return level_;
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::nBodies() const
-> size_t {

    return bodies_.size();
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::nCouplings() const
-> size_t {

    return couplings_.size();
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::leafVertexCounts() const
-> const std::vector<size_t>& {

    return leafVertexCounts_;
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::body(int i) const
-> const std::shared_ptr<Body>& {

    return bodies_[i];
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::deformedGrids() const
-> const std::vector<const DeformedGridType*>& {

    return deformedGrids_;
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::leafView(int bodyID) const
-> const DeformedLeafGridView& {

    return *leafViews_[bodyID];
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::levelViews(int bodyID) const
-> const std::vector<std::unique_ptr<const DeformedLevelGridView>>& {

    return levelViews_[bodyID];
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::levelView(int bodyID, int level) const
-> const DeformedLevelGridView& {

    return *levelViews_[bodyID][level];
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::leafVertexCount(int bodyID) const
-> int {

    return leafVertexCounts_[bodyID];
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::nBodyAssembler()
-> Dune::Contact::NBodyAssembler<DeformedGridType, Vector>& {

    return nBodyAssembler_;
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::nBodyAssembler() const
-> const Dune::Contact::NBodyAssembler<DeformedGridType, Vector>& {

    return nBodyAssembler_;
}

template <class GridType, int dimension>
auto LevelContactNetwork<GridType, dimension>::matrices() const
-> const Matrices<Matrix,2>& {

    return matrices_;
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::globalFriction(GlobalFrictionContainer& globalFriction) const {
    globalFriction.resize({nBodies()});

    for (size_t i=0; i<nBodies(); i++) {
        globalFriction[i] = bodies_[i]->globalFriction();
    }
}

template <class GridType, int dimension>
void LevelContactNetwork<GridType, dimension>::externalForces(ExternalForces& externalForces) const {
    externalForces.resize(nBodies());

    for (size_t i=0; i<nBodies(); i++) {
        externalForces[i] = &bodies_[i]->externalForce();
    }
}


#include "levelcontactnetwork_tmpl.cc"
