#ifndef SRC_DATA_STRUCTURES_LEVELCONTACTNETWORK_HH
#define SRC_DATA_STRUCTURES_LEVELCONTACTNETWORK_HH

#include <dune/common/fvector.hh>
#include <dune/common/bitsetvector.hh>

#include <dune/istl/bvector.hh>

#include <dune/contact/common/deformedcontinuacomplex.hh>
#include <dune/contact/common/couplingpair.hh>
#include <dune/contact/assemblers/nbodyassembler.hh>
//#include <dune/contact/assemblers/dualmortarcoupling.hh>
#include <dune/contact/projections/normalprojection.hh>

#include <dune/tectonic/bodydata.hh>

#include "../assemblers.hh"
#include "body.hh"
#include "enums.hh"
#include "globalfrictioncontainer.hh"
#include "matrices.hh"
//#include "multi-body-problem-data/myglobalfrictiondata.hh"

template <class GridType, int dimension> class LevelContactNetwork {
    private:
        using Vector = Dune::BlockVector<Dune::FieldVector<double, dimension>>;

    public:
        using Body = Body<GridType, Vector, dimension>;

        using DeformedGridType = typename Body::DeformedGridType;
        using DeformedLeafGridView = typename Body::DeformedLeafGridView;
        using DeformedLevelGridView = typename Body::DeformedLevelGridView;

        using DeformedLeafBoundaryPatch = BoundaryPatch<DeformedLeafGridView>;
        using DeformedLevelBoundaryPatch = BoundaryPatch<DeformedLevelGridView>;

        using Assembler = typename Body::Assembler;
        using Matrix = typename Assembler::Matrix;
        using LocalVector = typename Vector::block_type;
        using ScalarMatrix = typename Assembler::ScalarMatrix;
        using ScalarVector = typename Assembler::ScalarVector;
        using field_type = typename Matrix::field_type;
    
        using CouplingPair = Dune::Contact::CouplingPair<DeformedGridType, DeformedGridType, field_type>;
    
        using Function = Dune::VirtualFunction<double, double>;

        using BoundaryFunctions = std::vector<typename Body::BoundaryFunctions>;
        using BoundaryNodes = std::vector<typename Body::BoundaryNodes>;
        using BoundaryPatchNodes = std::vector<typename Body::BoundaryPatchNodes>;
        using BoundaryPatches = std::vector<typename Body::BoundaryPatches>;

        using GlobalFrictionContainer = GlobalFrictionContainer<typename Body::GlobalFriction, 2>;

        using StateEnergyNorms = std::vector<const typename Body::StateEnergyNorm *>;
        using ExternalForces = std::vector<const typename Body::ExternalForce *>;

        using NBodyAssembler = Dune::Contact::NBodyAssembler<DeformedGridType, Vector>;

    public:
        LevelContactNetwork(int nBodies, int nCouplings, int level = 0);

        void assemble();
        void assembleFriction(
                const Config::FrictionModel& frictionModel,
                const std::vector<ScalarVector>& weightedNormalStress);
		
        void setBodies(const std::vector<std::shared_ptr<Body>>& bodies);
        void setCouplings(const std::vector<std::shared_ptr<CouplingPair>>& couplings);

        void constructBody(
                const std::shared_ptr<BodyData<dimension>>& bodyData,
                const std::shared_ptr<GridType>& grid,
                std::shared_ptr<Body>& body) const;
        void constructCoupling(
                int nonMortarBodyIdx,
                int mortarBodyIdx,
                const std::shared_ptr<DeformedLevelBoundaryPatch>& nonmortarPatch,
                const std::shared_ptr<DeformedLevelBoundaryPatch>& mortarPatch,
                std::shared_ptr<CouplingPair>& coupling) const;

        // getter
        void totalNodes(
                const std::string& tag,
                Dune::BitSetVector<dimension>& totalNodes) const;
        void boundaryPatches(
                const std::string& tag,
                BoundaryPatches& patches) const;
        void boundaryPatchNodes(
                const std::string& tag,
                BoundaryPatchNodes& nodes) const;
        void boundaryNodes(
                const std::string& tag,
                BoundaryNodes& nodes) const;
        void boundaryFunctions(
                const std::string& tag,
                BoundaryFunctions& functions) const;

        void stateEnergyNorms(StateEnergyNorms& stateEnergyNorms) const;
        void globalFriction(GlobalFrictionContainer& globalFriction) const;
        void externalForces(ExternalForces& externalForces) const;

        auto level() const -> int;

        auto nBodies() const -> size_t;
        auto nCouplings() const -> size_t;

        auto body(int i) const -> const std::shared_ptr<Body>&;

        auto deformedGrids() const -> const std::vector<const DeformedGridType*>& ;

        auto leafVertexCounts() const -> const std::vector<size_t>&;
        auto leafVertexCount(int bodyID) const -> int;

        auto leafView(int bodyID) const -> const DeformedLeafGridView&;
        auto levelView(int bodyID, int level = 0) const -> const DeformedLevelGridView&;
        auto levelViews(int bodyID) const -> const std::vector<std::unique_ptr<const DeformedLevelGridView>>&;

        auto nBodyAssembler() -> Dune::Contact::NBodyAssembler<DeformedGridType, Vector>&;
        auto nBodyAssembler() const -> const Dune::Contact::NBodyAssembler<DeformedGridType, Vector>&;

        auto matrices() const -> const Matrices<Matrix,2>&;



    private:
        const int level_;

        std::vector<std::shared_ptr<Body>> bodies_;
        std::vector<std::shared_ptr<CouplingPair>> couplings_;

        std::vector<const DeformedGridType*> deformedGrids_;
        std::vector<std::unique_ptr<const DeformedLeafGridView>> leafViews_;
        std::vector<std::vector<std::unique_ptr<const DeformedLevelGridView>>> levelViews_;
        std::vector<size_t> leafVertexCounts_;

        NBodyAssembler nBodyAssembler_;

        Matrices<Matrix,2> matrices_;
};
#endif
